;* Sammlung diverser Funktionen/Unterprogramme in AVR-Assembler
;* Copyright: Freeware
;*
;History:
;20.5.2009	Erstellung
;27.8.2010	Fehler in milliwait0 korrigiert, Timer0-Aufruf auch fuer 328P.
;		Auch noch einfache Timer1-Variante eingefuegt
;		(Timer1-Aufrufe noch nicht fuer 328P angepasst)
;13.7.2014      Fehler in Divisionsroutinen divmodu16u8 und divmodu32u16
;               korrigiert: Ueberlauf abgefangen nach rol mit brcs
;               Quadratwurzel (sqrt), mulu8 angefuegt
;
; Liste der bisher vorhandenen Unterprogramme
; -------------------------------------------
;Mathematische Funktionen:
; Divisionen: divmod, divmodu16u8, divmodu16u16, divmodu32u16
; Multiplikationen: mulu16, mulu8
; Quadratwurzel: sqrt_8bit, sqrt_16bit
;Zeitbehandlungen:
; Timer-Uhrzeit: timer0_init, tc0over, timer1_init, timer1_match_A
; Wartefunktionen: milliwait
;
;-----------------------------------------------------------------------------
; Definitionsdatei fuer den Prozessortyp:
;.define ATMEGA328P
.ifdef ATMEGA328P
.include "m328Pdef.inc" ; ATmega328P
.else
.include "m8def.inc" 	; ATmega8
.endif

;-----------------------------------------------------------------------------
; divmod  15Bit Vorzeichenlose Division
; Eingabe: r3:r2 = Bit-Zahl (15Bit), r21:r20 = Teiler (16Bit)
; Ausgabe: r21:r20 = Resultat, r3:r2 = Rest
; Speicherverbrauch: 74 Byte
; Anzahl Takte:	33 - 393
;  Schnell wenn Zahl klein und Teiler gross ist
;  Teiler muss groesser 0 sein, sonst haengt das Programm
;-----------------------------------------------------------------------------
divmod:	push r4
	push r5
	push r16
	push r17
	ldi r16, 1
	ldi r17, 0		;z=1
	clr r4
	clr r5			;Vorlaeufiges Resultat=0
divL1:	cp r2, r20
	cpc r3, r21		;teiler>zahl ?
	brcs divend1		;ja-->
	lsl r20
	rol r21			;teiler*=2
	lsl r16
	rol r17			;z*=2
	rjmp divL1
divend1:		;teiler ist jetzt groesser als zahl
	cpi r16, 1	;ist teiler noch verdoppelt?
	breq divend2	;nein --> fertig
	lsr r21
	ror r20			;teiler/=2
	lsr r17
	ror r16			;z/=2
	cp r2, r20
	cpc r3, r21		;teiler>zahl ?
	brcs divend1		;ja-->
	sub r2, r20
	sbc r3, r21		;nein: zahl-=teiler
	add r4, r16
	adc r5, r17		;resultat += z
	rjmp divend1
divend2:mov r20, r4
	mov r21, r5		;r21:r20=Resultat  ;r3:r2=Rest
	pop r17
	pop r16
	pop r5
	pop r4
	ret

;-----------------------------------------------------------------------------
; divmodu16u8
;  16Bit vorzeichenlose Zahl dividieren durch 8Bit vorzeichenlosen Teiler
; Eingabe: r3:r2 = Zahl (16Bit), r0 = Teiler (8Bit)
; Ausgabe: r3:r2 = Resultat, r0 = Rest
; Speicherverbrauch: 36 Byte
; Anzahl Takte: 177 - 193
;  Wenn Teiler 0 ist wird das Resultat 0xFFFF sein
;-----------------------------------------------------------------------------
divmodu16u8:
	push r16
	push r17
	mov r16, r0
	clr r0
	ldi r17, 16
L1:	lsl r2
	rol r3
	rol r0
	brcs L3
	cp r0, r16
	brcs L2
L3:	sub r0, r16
	inc r2
L2:	dec r17
	brne L1
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; divmodu16u16
;  16Bit vorzeichenlose Zahl dividieren durch 16Bit vorzeichenlosen Teiler
; Eingabe: r3:r2 = Zahl, r1:r0 = Teiler
; Ausgabe: r3:r2 = Resultat, r1:r0 = Rest
; Speicherverbrauch: 48 Byte
; Anzahl Takte: 196 - 228
;  Wenn Teiler 0 ist wird das Resultat 0xFFFF sein
;-----------------------------------------------------------------------------
divmodu16u16:
	push r16
	push r17
	push r18
	mov r16, r0
	mov r17, r1
	clr r0
	clr r1
	ldi r18, 16
L3:	lsl r2
	rol r3
	rol r0
	rol r1
	cp r0, r16
	cpc r1, r17
	brcs L4
	sub r0, r16
	sbc r1, r17
	inc r2
L4:	dec r18
	brne L3
	pop r18
	pop r17
	pop r16
	ret

;-----------------------------------------------------------------------------
; mulu16
;  Vorzeichenlose Multiplikation zweier 16Bit-Zahlen
; Eingabe: r17:r16, r19:r18
; Ausgabe: r3:r2:r1:r0
mulu16:			;r3:r2:r1:r0 = r17:r16 * r19:r18
	push r4		;Abkuerzung fuer Kommentare:
	push r5		; AB=r17:r16, BC=r19:r18, EFGH=Ergebnis
	push r17
	mul r17, r19	;A*C
	mov r2, r0	; EFGH = AB*CD --> EF=A*C GH=B*D EFG+=A*D+B*C
	mov r3, r1	; r3:r2 = EF=A*C
	mul r16, r18	;B*D
	mov r4, r0
	mov r5, r1	; r5:r4 = GH=B*D
	mul r17, r18	;A*D
	clr r17
	add r5, r0	; EFG+=A*D
	adc r2, r1
	adc r3, r17
	mul r16, r19	;B*C
	add r5, r0	; EFG+=B*C
	adc r2, r1
	adc r3, r17
	mov r0, r4
	mov r1, r5
	pop r17
	pop r5
	pop r4
	ret

;-----------------------------------------------------------------------------
; noch nicht getestet: divmodu8  Division mit 8-Bit-Zahlen
divmodu8:			;r16/r17 --> Resultat=r16, Rest=r17
	push r18
	push r19
	clr r18
	ldi r19, 8
_L1:	lsl r16
	rol r18
	cp r18, r17
	brcs _L2
	sub r18, r17
	inc r16
_L2:	dec r19
	brne _L1
	mov r17, r18
	pop r19
	pop r18
	ret			;Anzahl Takte: 81 bis 89

divmodu8_leicht_optimiert:	;r16/r17 --> Resultat=r16, Rest=r17
	push r18
	push r19
	clr r18
	ldi r19, 8
_L1:	rol r16
	rol r18
	cp r18, r17
	brcs _L2
	sub r18, r17
_L2:	dec r19
	brne _L1
	rol r16
	com r16
	mov r17, r18
	pop r19
	pop r18
	ret			;Anzahl Takte: 83

divmodu8_opt:			;auf Geschwindigkeit optimiert (14 bis 54 Takte)
	cpi r17, 3		;Teiler<3 ?
	brcs _spezialfaelle	;ja-->
	push r18
	ldi r18, 0x80		;oberstes Bit als erstes inverses Resultatbit
	lsl r16
	rol r18
	rol r16
	rol r18			;um 2 Bits schieben bis erster Test
	cp r18, r17
	brcs _L1
	sub r18, r17
_L1:	rol r16			;Cy als inverses Resultatbit ins r16 schieben
	rol r18			;3 Bits geschoben
	cp r18, r17
	brcs _L2
	sub r18, r17
_L2:	rol r16
	rol r18
	cp r18, r17
	brcs _L3
	sub r18, r17
_L3:	rol r16
	rol r18
	cp r18, r17
	brcs _L4
	sub r18, r17
_L4:	rol r16
	rol r18
	cp r18, r17
	brcs _L5
	sub r18, r17
_L5:	rol r16
	rol r18
	cp r18, r17
	brcs _L6
	sub r18, r17
_L6:	rol r16
	rol r18			;8 Bits geschoben
	cp r18, r17
	brcs _L7
	sub r18, r17
_L7:	rol r16			;letztes Resultatbit schieben
	com r16			;Bits im Resultat umkehren
	mov r17, r18		;r17=Rest
	pop r18
	ret		;54 Takte
_spezialfaelle:			;Teilen durch 0, 1, oder 2 als Spezialfaelle
	cpi r17, 1
	ldi r17, 0		;r17=Rest=0
	breq _sp1
	brcs _sp0
	lsr r16			;Division durch 2
	rol r17			;r17=Rest
	ret		;16 Takte mit Teiler 2
_sp0:	ldi r16, 0xFF		;Maximalwert bei Division durch 0
_sp1:	ret		;14 Takte mit Teiler 1, 16 Takte mit Teiler 0

;-----------------------------------------------------------------------------
; noch nicht getestet: div10  Division durch 10
div10_gross:			;r16/10 --> Resultat=r16, Rest=r17
	push r18
	push r19
	clr r17
	lsl r16
	rol r17
	lsl r16
	rol r17
	lsl r16
	rol r17
	ldi r18, 10
	ldi r19, 5		;noch 5 mal schieben
_L1:	lsl r16
	rol r17
	cp r17, r18
	brcs _L2
	sub r18, r17
	inc r16
_L2:	dec r19
	brne _L1
	pop r19
	pop r18
	ret			;Anzahl Takte: 73 bis 78
div10_klein:
	mov r17, r16
	ldi r16, -1
_L1:	inc r16
	subi r17, 10
	brcc _L1
	subi r17, -10
	ret			;Anzahl Takte: 13 bis 113
div10_mittel:
	mov r17, r16
	ldi r16, -7
_L1:	subi r16, -6
	subi r17, 60
	brcc _L1
	subi r17, -60
_L2:	inc r16
	subi r17, 10
	brcc _L2
	subi r17, -10
	ret			;Anzahl Takte: 17 bis 41
div10:
	mov r17, r16
	clr r16
	cpi r17, 160
	brcs _L1
	ldi r16, 16
	subi r17, 160
_L1:	cpi r17, 80
	brcs _L2
	ori r16, 8
	subi r17, 80
_L2:	cpi r17, 40
	brcs _L3
	ori r16, 4
	subi r17, 40
_L3:	cpi r17, 20
	brcs _L4
	ori r16, 2
	subi r17, 20
_L4:	cpi r17, 10
	brcs _L5
	inc r16
	subi r17, 10
_L5:	ret			;Anzahl Takte: 24 bis 28

;-----------------------------------------------------------------------------
; mulu8
;  Vorzeichenlose Multiplikation einer 16Bit-Zahl mit einer 8Bit-Zahl
; Eingabe: r17:r16, r18  (AB*C --> FGH)
; Resultat: r18:r17:r16, zerstoert: r0,r1
mulu8:	mul r16, r18	;B*C
	mov r16, r0	;r1:r16 = B*C
	push r1
	mul r17, r18	;r1:r0 = A*C
	pop r17		;r17:r16 = B*C
	ldi r18, 0	; F=0
	add r17, r0	; FG+=A*C
	adc r18, r1
	ret

;-----------------------------------------------------------------------------
; divmodu32u16
; Vorzeichenlose Division einer 32Bit-Zahl durch eine 16Bit-Zahl
; Eingabe: r3:r2:r1:r0 / r17:r16
; Ausgabe: r3:r2:r1:r0 = Ergebnis, r17:r16 = Rest
divmodu32u16:			;r3:r2:r1:r0 /= r17:r16, r17:r16 = Rest
	push r18
	push r19
	push r20
	clr r18
	clr r19
	ldi r20, 32
div16L1:lsl r0
	rol r1
	rol r2
	rol r3
	rol r18
	rol r19
	brcs div16L3
	cp  r18, r16
	cpc r19, r17
	brcs div16L2
div16L3:sub r18, r16
	sbc r19, r17
	inc r0
div16L2:dec r20
	brne div16L1
	mov r16, r18
	mov r17, r19		;r17:r16 = Rest
	pop r20
	pop r19
	pop r18
	ret

;-----------------------------------------------------------------------------
; divur20  Auf Geschwindigkeit Optimierte Variante fuer Wurzelberechnung, Minimaler Teiler: 16
; Anzahl Takte: 140 bis 152
divur20:		;r17:r16 / r20 --> r19:r18, r0=Rest, unveraendert: r17:r16, zerstoert: r1
	ldi r18, 12		;Es muessen noch 12 Bits in der Schlaufe behandelt werden
	mov r1, r18		;r1 = Schlaufenzaehler
	ldi r18, 0x0F		;Erste 4 Bits sofort schieben, da minimaler Teiler=16 ist
	mov r19, r17
	swap r19
	mov r0, r19
	and r0, r18		;Oberste 4 Bits von r17:r16 ins r0 geschoben
	andi r19, 0xF0		;Naechste 4 Bits in r19 nach links geschoben
	swap r16
	and r18, r16		;obere 4 Bits von r16 in untere 4 Bits von r19 kopieren
	or  r19, r18
	mov r18, r16
	andi r18, 0xF0		;unterste 4 Bits in r18 nach links geschoben
	swap r16		;r16 unveraendert
_L1:	lsl r18
	rol r19
	rol r0
	brcs _L3
	cp r0, r20
	brcs _L2
_L3:	sub r0, r20
	inc r18			;Resultat-Bit setzen
_L2:	dec r1
	brne _L1
	ret

;-----------------------------------------------------------------------------
; Quadratwurzel

sqrt_8bit_variante1:			;Wurzel von r16 berechnen, Resultat: r16
	push ZL
	push ZH
	ldi ZL, low(sqrt8_tabelle<<1)
	ldi ZH, high(sqrt8_tabelle<<1)
	add ZL, r16
	ldi r16, 0
	adc ZH, r16
	lpm r16, Z
	pop ZH
	pop ZL
	ret			;22 Takte
sqrt8_tabelle:
	.db 0,1,1,2,2,2,2,3,3,3,3,3,3,4,4,4 ;0...15
	.db 4,4,4,4,4,5,5,5,5,5,5,5,5,5,5,6 ;16...31
	.db 6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7 ;32...47
	.db 7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8 ;48...63
	.db 8,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9 ;64...79
	.db 9,9,9,9,9,9,9,9,9,9,9,10,10,10,10,10 ;80...95
	.db 10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,11 ;96...111
	.db 11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11 ;112...127
	.db 11,11,11,11,11,12,12,12,12,12,12,12,12,12,12,12 ;128...143
	.db 12,12,12,12,12,12,12,12,12,12,12,12,12,13,13,13 ;144...159
	.db 13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13 ;160...163
	.db 13,13,13,13,13,13,13,14,14,14,14,14,14,14,14,14 ;176...191
	.db 14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14 ;192...207
	.db 14,14,14,15,15,15,15,15,15,15,15,15,15,15,15,15 ;208...223
	.db 15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15 ;224...239
	.db 15,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16 ;240...255

;; Komplizierter, aber schneller und braucht weniger Flash-Speicher! (128 Bytes)
;; noch testen
sqrt_8bit:			;Wurzel von r16 berechnen, Resultat: r16
	cpi r16, 56+1		;x>56
	brcs _L56		;nein-->
	cpi r16, 132+1		;x>132
	brcs _L132
	cpi r16, 182+1
	brcs _L182
	cpi r16, 210+1
	brcs _L210
	cpi r16, 240+1		;x>240
	brcc _res16		;ja --> Resultat=16
	ldi r16, 15		;nein: x>210 --> Resultat=15
	ret			;18 Takte
_res16:	ldi r16, 16
	ret			;19 Takte
_L210:	ldi r16, 14
	ret			;17 Takte
_L182:	cpi r16, 156+1		;x>156
	brcc _res13		;ja --> Resultat=13
	ldi r16, 12		;nein: Resultat=12
	ret			;17 Takte
_res13:	ldi r16, 13
	ret			;18 Takte
_L132:	cpi r16, 90+1
	brcs _L90
	cpi r16, 110+1		;x>110
	brcc _res11		;ja-->
	ldi r16, 10
	ret
_res11:	ldi r16, 11
	ret
_L90:	cpi r16, 72+1		;x>72
	brcc _res9		;ja-->
	ldi r16, 8
	ret
_res9:	ldi r16, 9
	ret			;19 Takte
_L56:				;x<=56
	cpi r16, 12+1
	brcs _L12
	cpi r16, 30+1
	brcs _L30
	cpi r16, 42+1		;x>42
	brcc _res7		;ja-->
	ldi r16, 6
	ret
_res7:	ldi r16, 7
	ret
_L30:	cpi r16, 20+1		;x>20
	brcc _res5		;ja-->
	ldi r16, 4
	ret
_res5:	ldi r16, 5
	ret
_L12:	cpi r16, 2+1
	brcs _L1
	cpi r16, 6+1		;x>6
	brcc _res3		;ja-->
	ldi r16, 2
	ret
_res3:	ldi r16, 3
	ret			;19 Takte
_L1:	tst r16
	breq _L0
	ldi r16, 1
_L0:	ret			;19 Takte

sqrt_16bit_variante0:		;Wurzel von r17:r16 berechnen, Resultat: r16, zerstoert: r0-r1
	push r20		;sehr kurz, aber langsam und ungenau
	ldi r20, 255
_L1:	mul r20, r20
	cp  r16, r0
	cpc r17, r1
	brcc _fertig
	dec r20
	brne _L1
_fertig:	;; Resultat ist eventuell abgerundet
	mov r16, r20
	pop r20
	ret			;maximal 2052 Takte

sqrt_16bit_variante1:		;Wurzel von r17:r16 berechnen, Resultat: r16, zerstoert: r0-r1
	push r18		;kurz aber langsam
	push r19
	push r20
	ldi r20, 255
	mul r20, r20
	sub r0, r16
	sbc r1, r17
	breq _fertig
	brcs _fertig
_L1:	movw r18, r0		;Differenz in r19:r18 speichern
	dec r20
_L2:	mul r20, r20
	sub r0, r16
	sbc r1, r17
	breq _fertig
	brcc _L1
	movw r16, r0		;Vorherige Differenz in r17:r16 speichern
	com r16			;Wert negieren mit Zweiercomplement:
	com r17			;Bits umkehren,
	subi r16, -1		;dann 1 addieren
	sbci r17, -1
	cp  r18, r16		;vorheriger Wert kleinere Abweichung?
	cpc r19, r17
	brcc _fertig		;nein-->
	inc r20			;ja: vorheriger Wert nehmen
_fertig:mov r16, r20
	pop r20
	pop r19
	pop r18
	ret			;maximal 2326 Takte

.def zero=r5	;weglassen wenn schon im Hauptprogramm gemacht

sqrt_16bit_variante2:		;Wurzel von r17:r16 berechnen, Resultat: r16, zerstoert: r0-r1, (r5=zero)
	tst r17			;Oberes Byte leer?
	breq sqrt_8bit		;ja--> Wurzel aus 8-Bit-Zahl
	push r18		;Heron-Verfahren fuer Wurzel(y): Schaetzwert a, b=y/a, neues a=(a+b)/2
	push r19		;Hier: y-Minimalwert=256, y-Maximalwert=65535 
	push r20		;      beim Maximalwert wird das Resultat 255 sein, also abgerundeter Wert
	clr zero	;weglassen wenn schon im Hauptprogramm gemacht
	ldi r18, high(64771)
	cpi r16, low(64771)	;x>=64771
	cpc r17, r18
	brcc _maxwert		;ja: Ergebnis=Maximalwert=255
	ldi r18, 16		;Startwert 16
	mov r19, r17
	rjmp _L0
_maxwert:
	ldi r20, 255
	rjmp _fertig
_max_startwert:
	ldi r18, 255
	rjmp _L1
_startwert:
	lsl r18			;besserer Startwert
	brcs _max_startwert
	lsr r19
_L0:	lsr r19
	brne _startwert
_L1:	mov r20, r18		;r20 = a
	rcall divur20		;r17:r16 / r20 --> r19:r18, r19:r18 = b
	lsl r0
	cp  r20, r0		;Rest>a/2 ? ja: Cy setzen und somit aufrunden
	adc r18, r20		;a+b + Cy fuer Rundung
	adc r19, zero		;a+b
	lsr r19
	brne _L2
	ror r18			;(a+b)/2
	rjmp _L3
_L2:	ldi r18, 255		;Maximalwert
_L3:	cp r18, r20		;neues a==vorheriges a?
	breq _fastfertig
	dec r20
	cp r18, r20		;neues a nur um 1 verschieden zu verherigem?
	brne _L1		;nein-->
_fastfertig:
	mul r20, r20		;a*a
	movw r18, r16
	sub r18, r0
	sbc r19, r1		;r19:r18 = x-a*a, a*a<x ?
	breq _fertig		;nein-->
	brcs _fertig		;nein-->
	inc r20			;r20=a+1
	mul r20, r20		;(a+1)*(a+1)
	sub r0, r16
	sbc r1, r17		;(a+1)*(a+1)-x
	cp  r0, r18
	cpc r1, r19		;(a+1)*(a+1)-x < x-a*a ?
	brcs _fertig		;ja: Resultat=a+1
	dec r20			;nein: Resultat=a
_fertig:
	mov r16, r20
	pop r20
	pop r19
	pop r18
	ret

;; variante3: bisher beste Variante:
sqrt_16bit_variante3:		;Wurzel von r17:r16 berechnen, Resultat: r16, zerstoert: r0-r1
	tst r17			;Oberes Byte leer?
	breq sqrt_8bit		;ja--> Wurzel aus 8-Bit-Zahl
	push r18
	push r19
	push r20
	ldi r20, 0x80		;a=0x80=Startwert
	cpi r17, 0x40		;x>=a*a ?
	brcc _start		;ja--> starten
	ldi r18, 0x40		;0x4000 entspricht a*a
_L0:	lsr r20			;besserer Startwert berechnen
	lsr r18
	lsr r18
	cp  r17, r18		;x>=a*a ?
	brcs _L0		;nein-->
_start:	mov r19, r20		;mask setzen
_L1:	lsr r19
	breq _fastfertig
_L2:	mov r18, r20
	or  r20, r19		;Bit setzen
	mul r20, r20
	cp  r16, r0
	cpc r17, r1		;a*a>=x ?
	breq _fertig		;gleich --> schon fertig
	brcc _L1		;kleiner -->
	mov r20, r18		;groesser: Bit wieder loeschen
	lsr r19
	brne _L2
_fastfertig:
	cpi r20, 255
	breq _fertig
	mul r20, r20		;a*a
	movw r18, r16
	sub r18, r0
	sbc r19, r1		;r19:r18 = x-a*a
	inc r20			;r20=a+1
	mul r20, r20		;(a+1)*(a+1)
	sub r0, r16
	sbc r1, r17		;(a+1)*(a+1)-x
	cp  r0, r18
	cpc r1, r19		;(a+1)*(a+1)-x < x-a*a ?
	brcs _fertig		;ja: Resultat=a+1
	dec r20			;nein: Resultat=a
_fertig:
	mov r16, r20
	pop r20
	pop r19
	pop r18
	ret			;max 128 Takte

;; variante4 noch testen, sollte schnellste Variante sein
sqrt_16bit:			;Wurzel von r17:r16 berechnen, Resultat: r16, zerstoert: r0-r1
	tst r17			;Oberes Byte leer?
	breq sqrt_8bit		;ja--> Wurzel aus 8-Bit-Zahl
	push r18
	push r19
	push r20
	cpi r17, 0x10		;x>=0x40*0x40 ?
	brcs _s1		;nein-->
	ldi r20, 0x80		;a=Startwert
	cpi r17, 0x40		;x>=0x80*0x80 ?
	brcc _start		;ja-->
	rjmp _s2
_s1:	ldi r20, 0x20		;a=Startwert
	cpi r17, 0x04		;x>=0x20*0x20 ?
	brcc _start		;ja-->
_s2:	lsr r20			;nein: kleinerer Startwert
_start:	mov r19, r20		;mask setzen
_L1:	lsr r19
	breq _fastfertig
_L2:	mov r18, r20
	or  r20, r19		;Bit setzen
	mul r20, r20
	cp  r16, r0
	cpc r17, r1		;a*a>=x ?
	breq _fertig		;gleich --> schon fertig
	brcc _L1		;kleiner -->
	mov r20, r18		;groesser: Bit wieder loeschen
	lsr r19
	brne _L2
_fastfertig:
	cpi r20, 255		;Maximalwert?
	breq _fertig		;ja-->
	mul r20, r20		;a*a
	movw r18, r16
	sub r18, r0
	sbc r19, r1		;r19:r18 = x-a*a
	inc r20			;r20=a+1
	mul r20, r20		;(a+1)*(a+1)
	sub r0, r16
	sbc r1, r17		;(a+1)*(a+1)-x
	cp  r0, r18
	cpc r1, r19		;(a+1)*(a+1)-x < x-a*a ?
	brcs _fertig		;ja: Resultat=a+1
	dec r20			;nein: Resultat=a
_fertig:
	mov r16, r20
	pop r20
	pop r19
	pop r18
	ret			;max 127 Takte

;; Optimierungen in sqrt_16bit noch nicht getestet.
teste_wurzelprogramm:		;r16-r17 zerstoert
	push r0
	push r1
	push r18
	push r19
	ldi r18, 0
	ldi r19, 0		;r19:r18=Startwert=0
_L1:	movw r16, r18
	rcall sqrt_16bit	;Wurzel berechnen
	rcall uart_putc		;Resultat ueber Serielle Schnittstelle schicken
	subi r18, -1
	sbci r19, -1		;naechster Wert
	brne _L1
	pop r19
	pop r18
	pop r1
	pop r0
	ret

teste_wurzel_varianten:		;r16-r17 zerstoert
	push r0
	push r1
	push r18
	push r19
	push r20
	ldi r18, 0
	ldi r19, 0		;r19:r18=Startwert=0
_L1:	movw r16, r18
	rcall sqrt_16bit	;Wurzel berechnen
	mov r20, r16
	movw r16, r18
	rcall sqrt_16bit_variante1	;Wurzel anders berechnen
	cpse r16, r20			;Resultate gleich?
	rcall wurzel_fehler		;nein: Fehlermeldung
	rcall uart_putc		;Resultat ueber Serielle Schnittstelle schicken
	subi r18, -1
	sbci r19, -1		;naechster Wert
	brne _L1
	pop r20
	pop r19
	pop r18
	pop r1
	pop r0
	ret

wurzel_fehler:
	sbi LED_PORT, LED_BIT	;Fehler-LED einschalten
	ret

;-----------------------------------------------------------------------------
; Timer-Uhr
; Registerbelegungen:
; r11:r10 Millisekunden, r12 Sekunden, r13 Minuten, r14 Stunden, r15 Tage
.def millisecL=r10
.def millisecH=r11
.def sec=r12
.def min=r13
.def stunden=r14
.def tage=r15

;-----------------------------------------------------------------------------
; Variante mit Timer0:
; Einschraenkungen: ist nur mit runder Anzahl MHz genau.
;	rjmp tc0over		; 10  TC0 Overflow
.equ	ZAEHLER =131	;131 bei 8MHz, 190 bei 4.19MHz 198 oder 199 bei 3.68MHz
.ifdef ATMEGA328P
timer0_init:
	push r17
	ldi r17, 0x01
	sts TIMSK0, r17		;Timer/Counter0 Overflow einschalten
	ldi r17, 3		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0B, r17		;Vorteiler auf 64 setzen
	ldi r17, ZAEHLER
	out TCNT0, r17
	pop r17
	ret
.else
timer0_init:
	push r17
	ldi r17, 0x01
	out TIMSK, r17		;Timer/Counter0 Overflow einschalten
	ldi r17, 3		;1=nicht teilen, 2=durch 8, 3=64, 4=256, 5=1024
	out TCCR0, r17		;Vorteiler auf 64 setzen
	ldi r17, ZAEHLER
	out TCNT0, r17
	pop r17
	ret
.endif
; Timer0-Interrupt-Routine
tc0over:push r16
	push r17
	in r17, SREG	;Statusregister sichern
	push r17
	ldi r17, ZAEHLER
	out TCNT0, r17
	ldi r16, 0xE8
	ldi r17, 0x03	;r17:r16 = 1000 = 0x03E8
	inc r10		;Millisekunden erhoehen
	brne L1
	inc r11
L1:	cp r10, r16
	brne end1	;fertig wenn msec!=1000
	cp r11, r17
	brne end1	;fertig wenn msec!=1000
	clr r10
	clr r11		;Millisekunden auf 0 setzen
	inc r12		;Sekunden erhoehen
	ldi r16, 60
	cp r12, r16
	brne end1	;fertig wenn sec!=60
	clr r12		;Sekunden auf 0 setzen
	inc r13		;Minuten erhoehen
	cp r13, r16
	brne end1	;fertig wenn min!=60
	clr r13		;Minuten auf 0 setzen
	inc r14		;Stunden erhoehen
	ldi r16, 24
	cp r14, r16
	brne end1	;fertig wenn std!=24
	clr r14		;Stunden auf 0 setzen
	inc r15		;Tage erhoehen
end1:	pop r17
	out SREG, r17	;Statusregister zurueckholen
	pop r17
	pop r16
	reti

;-----------------------------------------------------------------------------
; Einfache Variante mit Timer1:
; Laeuft mit beliebiger Anzahl MHz sehr genau.
;	rjmp timer1_match_A	; 7   TC1 Compare Match A

; Quarz aufs Hertz genau ausmessen und hier eintragen
.equ Hz=3686400			; exakt 3.6864MHz

;//Ausmessung-Beispiel: 3.6864MHz ist nach 20 Stunden 0.35 sec nachgegangen:
;// 3686400 / (3600*20) * 0.35 = 17.92  3686400-35.84=3686364.16
;.equ Hz=3686364

.equ Rest1 = Hz-Hz/1000*1000	;//jede Sekunde zu korrigierender Fehler
.equ NormalerCompwert = Hz/1000-1

timer1_init:
	push r16
	push r17
	ldi r16, (1<<WGM12)|(1<<CS10)	;CTC-OCR1A-Modus, Prescaler=1
	out TCCR1B, r16
	ldi r16, low(NormalerCompwert)
	ldi r17, high(Normalercompwert)
	out OCR1AH, r17
	out OCR1AL, r16
	ldi r16, 0		;Startwert des Timers = 0
	out TCNT1H, r16
	out TCNT1L, r16
	ldi r16, 1<<OCIE1A	;Timer1 Interrupts bei Vergleichswert
	out TIMSK, r16
	clr millisecH
	clr millisecL
	clr sec
	clr min
	clr stunden
	clr tage
	pop r17
	pop r16
	ret

timer1_match_A:
	push r16
	push r17
	inc millisecL		;++ms
	brne tc1_L1
	inc millisecH
tc1_L1: ldi r16, low(1000)
	ldi r17, high(1000)
	cp  millisecL, r16
	cpc millisecH, r17
	breq tc1_sec
	ldi r16, low(1000-Rest1)
	ldi r17, high(1000-Rest1)
	cp  millisecL, r16
	cpc millisecH, r17	;millisec>=1000-Rest1 ?
	brcs tc1_fertig		;nein-->
	ldi r16, low(NormalerCompwert+1)
	ldi r17, high(Normalercompwert+1)
	rjmp tc1_L2
tc1_fertig:
	ldi r16, low(NormalerCompwert)
	ldi r17, high(Normalercompwert)
tc1_L2:	out OCR1AH, r17
	out OCR1AL, r16
	pop r17
	pop r16
	reti
tc1_sec:			; Sekunden erhoehen
	clr millisecL
	clr millisecH
	inc sec
	ldi r17, 60
	cp  sec, r17		;sec==60?
	brne tc1_fertig		;nein-->
	clr sec
	inc min
	cp min, r17
	brne tc1_fertig
	clr min
	inc stunden
	ldi r17, 24
	cp stunden, r17
	brne tc1_fertig
	clr stunden
	inc tage		;{stunden=0; ++tage;}
	rjmp tc1_fertig

;-----------------------------------------------------------------------------
; Hochgenaue Variante mit Timer1:
; Laeuft mit beliebiger Anzahl MHz sehr genau.
;	rjmp timer1_match_A	; 7   TC1 Compare Match A

; Quarz moeglichst genau ausmessen und hier eintragen
; in Hertz und Tausendstel Hertz:

.equ Hz=3686400			; exakt 3.6864MHz
.equ MilliHz=0			; 3'686'400.000 Hz

;//Ausmessung-Beispiel: 3.6864MHz ist nach 20 Stunden 0.35 sec nachgegangen:
;// 3686400 / (3600*20) * 0.35 = 17.92  3686400-17.92=3686382.080
;.equ Hz=3686382
;.equ MilliHz=80

.equ Rest1 = Hz-Hz/1000*1000	;//jede Sekunde zu korrigierender Fehler
.equ Rest2 = MilliHz*6/100	;//jede Minute
.equ Rest3 = ((MilliHz*6-Rest2*100)*60+50)/100 ;//jede Stunde
.equ NormalerCompwert = Hz/1000-1

timer1_init:
	push r16
	push r17
	ldi r16, (1<<WGM12)|(1<<CS10)	;CTC-OCR1A-Modus, Prescaler=1
	out TCCR1B, r16
	ldi r16, low(NormalerCompwert)
	ldi r17, high(Normalercompwert)
	out OCR1AH, r17
	out OCR1AL, r16
	ldi r16, 0		;Startwert des Timers = 0
	out TCNT1H, r16
	out TCNT1L, r16
	ldi r16, 1<<OCIE1A	;Timer1 Interrupts bei Vergleichswert
	out TIMSK, r16
	clr millisecH
	clr millisecL
	clr sec
	clr min
	clr stunden
	clr tage
	pop r17
	pop r16
	ret

timer1_match_A:
	push r16
	push r17
	push r18		;uchar mehr;
	clr r18			;mehr=0
	inc millisecL		;++ms
	brne tc1_L1
	inc millisecH
tc1_L1: ldi r16, low(1000)
	ldi r17, high(1000)
	cp  millisecL, r16
	cpc millisecH, r17
	breq tc1_sec
	ldi r16, low(1000-Rest1)
	ldi r17, high(1000-Rest1)
	cp  millisecL, r16
	cpc millisecH, r17	;millisec>=1000-Rest1 ?
	brcs tc1_fertig		;nein-->
	ldi r18, 1		;ja: mehr=1
tc1_fertig:
	ldi r16, low(NormalerCompwert)
	ldi r17, high(Normalercompwert)
	add r16, r18
	brcc tc1_L2
	inc r17
tc1_L2:	out OCR1AH, r17
	out OCR1AL, r16
	pop r18
	pop r17
	pop r16
	reti
tc1_sec:			; Sekunden erhoehen
	clr millisecL
	clr millisecH
	inc sec
	ldi r17, 60
.if Rest2!=0
	ldi r16, 60-Rest2
	cp sec, r16		;if(++sec>=60-Rest2) mehr=1; else mehr=0;
	brcs tc1_L3
	ldi r18, 1
.endif
tc1_L3:	cp  sec, r17		;sec==60?
	brne tc1_fertig		;nein-->
	clr sec
	clr r18
	inc min
.if Rest3!=0
	ldi r16, 60-Rest3
	cp min, r16		;if(++min>=60-Rest3) mehr=1; else mehr=0;
	brcs tc1_L4
	ldi r18, 1
.endif
tc1_L4:	cp min, r17
	brne tc1_fertig
	clr min
	clr r18
	inc stunden
	ldi r17, 24
	cp stunden, r17
	brne tc1_fertig
	clr stunden
	inc tage		;{stunden=0; ++tage;}
	rjmp tc1_fertig

;-----------------------------------------------------------------------------
; Wartefunktionen
;nicht optimierte Variante
milliwait0:                      ;Wartezeit: r17:r16 msec, min 1 max 999
        push r16
        push r17
        push r18
        push r19
	cli
        mov r18, r10
        mov r19, r11            ;r19r18 = sollzeit = aktuelle Millisec
	sei
        add r18, r16
        adc r19, r17            ;sollzeit += Wartezeit
	ldi r16, low(1000)
        ldi r17, high(1000)
        cp  r18, r16
	cpc r19, r17		;auf 1000 testen
        brcs loop1              ;sollzeit<1000 --> weiter
	sub r18, r16		;sonst sollzeit>=1000 --> subtrahiere 1000
        sbc r19, r17
loop1:	cp r18, r10
        cpc r19, r11	;auf sollzeit warten
        brne loop1
	cp r18, r10	;falls zwischendurch ein Interrupt war:
        brne loop1	;koennte z.B. von 0x00FF auf 0x0100 gewechselt
        pop r19		;haben waerend wir auf 0x01FF warten.
        pop r18
        pop r17
        pop r16
        ret

;-----------------------------------------------------------------------------
; milliwait  optimierte Variante  Wartezeit 1 bis 1000 Millisekunden
; in r11:r10 wird durch einen Timerinterrupt die aktuelle Millisekunde
; gespeichert (16Bit-Zahl von 0 bis 999).
; Die aktelle Sekunde ist in r12 (wird hier noch nicht benutzt, da nur
; bis maximal 1000ms gewartet werden soll).
milliwait1:			;Wartezeit: r16, 1 bis 255 msec
	push r17
	clr r17
	rjmp miL1
milliwait:			;Wartezeit: r17:r16 msec, 1 bis 1000
	push r17
miL1:	push r16
	push r18
	mov r18, r10
milop1:	cp  r18, r10
	breq milop1		;1. Millisekunde abwarten (0 bis 1 ms)
	add r16, r10
	adc r17, r11		;sollzeit = Wartezeit + aktuelle Millisek
	subi r16, low(1001)
	sbci r17, high(1001)	;sollzeit -= 1; sollzeit -= 1000;
	brpl milop2		;wenn sollzeit>=0 -->
	subi r16, low(-1000)	;sonst haben wir 1000 zuviel subtrahiert
	sbci r17, high(-1000)	;also wieder 1000 addieren (soll -= -1000)
milop2:	cp r16, r10
	cpc r17, r11		;auf sollzeit warten
	brne milop2
	cp r16, r10		;falls zwischen 1. u 2. cp ein Interrupt war
	brne milop2
	pop r18
	pop r16
	pop r17
	ret

;-----------------------------------------------------------------------------
;; UART Routinen
.equ F_CPU = 8000000                            ; Systemtakt in Hz
.equ BAUD  = 9600                               ; Baudrate
 
; Berechnungen
.equ UBRR_VAL   = ((F_CPU+BAUD*8)/(BAUD*16)-1)  ; clever runden
.equ BAUD_REAL  = (F_CPU/(16*(UBRR_VAL+1)))     ; Reale Baudrate
.equ BAUD_ERROR = ((BAUD_REAL*1000)/BAUD-1000)  ; Fehler in Promille
 
.if ((BAUD_ERROR>10) || (BAUD_ERROR<-10))       ; max. +/-10 Promille Fehler
  .error "Systematischer Fehler der Baudrate groesser 1 Prozent und damit zu hoch!"
.endif

uart_init:
	ldi r16, UBRR_VAL>>8
	out UBRRH, r16
	ldi r16, UBRR_VAL&0xFF
	out UBRRL, r16
	ldi r16, (1<<RXEN)|(1<<TXEN)
	out UCSRB, r16
;	ldi r16, (1<<USBS)|(3<<UCSZ0) ;8N2 ATtiny2313A
	ldi r16, (1<<URSEL)|(3<<UCSZ0) ;8N1 Atmega8
;	ldi r16, (3<<UCSZ0) ;8N1 ATtiny2313A
	out UCSRC, r16
	ret

uart_putc:			; zu sendendes Zeichen in r16
	sbis UCSRA, UDRE
	rjmp uart_putc
	out UDR, r16
	ret

uart_getc:			; Resultat in r16
	sbis UCSRA, RXC
	rjmp uart_getc
	in r16, UDR
	ret
;-----------------------------------------------------------------------------
